import { range, descending } from 'd3-array';
import React, { useRef, useState, useEffect, useMemo } from 'react';
import seedrandom from 'seedrandom';
import { select, event } from 'd3-selection';
import ResizeObserver from 'resize-observer-polyfill';
import 'd3-transition';
import tippy from 'tippy.js';
import { scaleOrdinal } from 'd3-scale';
import { schemeCategory10 } from 'd3-scale-chromatic';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

var __assign = function() {
    __assign = Object.assign || function __assign(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};

function __spreadArrays() {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
}

// @ts-nocheck
// Word cloud layout by Jason Davies, https://www.jasondavies.com/wordcloud/
// Algorithm due to Jonathan Feinberg, http://static.mrfeinberg.com/bv_ch03.pdf
var dispatch = require("d3-dispatch").dispatch;
var cloudRadians = Math.PI / 180, cw = 1 << 11 >> 5, ch = 1 << 11;
function cloud () {
    var size = [256, 256], text = cloudText, font = cloudFont, fontSize = cloudFontSize, fontStyle = cloudFontNormal, fontWeight = cloudFontNormal, rotate = cloudRotate, padding = cloudPadding, spiral = archimedeanSpiral, words = [], timeInterval = Infinity, stepLimit = Infinity, event = dispatch("word", "end"), timer = null, random = Math.random, cloud = {}, canvas = cloudCanvas, maxSizeToFit;
    cloud.canvas = function (_) {
        return arguments.length ? (canvas = functor(_), cloud) : canvas;
    };
    cloud.start = function () {
        var contextAndRatio = getContext(canvas()), board = zeroArray((size[0] >> 5) * size[1]), bounds = null, n = words.length, i = -1, tags = [], data = words.map(function (d, i) {
            d.text = text.call(this, d, i);
            d.font = font.call(this, d, i);
            d.style = fontStyle.call(this, d, i);
            d.weight = fontWeight.call(this, d, i);
            d.rotate = rotate.call(this, d, i);
            d.size = ~~fontSize.call(this, d, i);
            d.padding = padding.call(this, d, i);
            return d;
        }).sort(function (a, b) { return b.size - a.size; });
        var stepStart = +new Date();
        if (timer)
            clearInterval(timer);
        timer = setInterval(step, 0);
        setTimeout(function () { return step(); }, 0);
        return cloud;
        function step() {
            if (size[0] === 0 || size[1] === 0) {
                cloud.stop();
                event.call('end', cloud, tags, bounds);
                return;
            }
            var start = +new Date, d, x, y, nsi, shifted, decremented;
            while (+new Date - start < timeInterval && ++i < n && timer) {
                d = data[i];
                x = d.x = (size[0]) >> 1;
                y = d.y = (size[1]) >> 1;
                nsi = i + 1;
                cloudSprite(contextAndRatio, d, data, i);
                while (d.hasText) {
                    if (place(board, d, bounds)) {
                        // it fits!
                        maxSizeToFit = d.size;
                        tags.push(d);
                        event.call("word", cloud, d);
                        if (bounds)
                            cloudBounds(bounds, d);
                        else
                            bounds = [
                                { x: d.x + d.x0, y: d.y + d.y0 },
                                { x: d.x + d.x1, y: d.y + d.y1 }
                            ];
                        // Temporary hack
                        d.x -= size[0] >> 1;
                        d.y -= size[1] >> 1;
                        break;
                    }
                    else {
                        // reset
                        delete d.sprite;
                        d.sprite = null;
                        d.x = x;
                        d.y = y;
                        // decrement the size and try again
                        if (maxSizeToFit < d.size) {
                            d.size = maxSizeToFit;
                        }
                        else if (data[nsi] === undefined) {
                            break;
                        }
                        else if ((nsi <= data.length) && (data[nsi].size < d.size)) {
                            d.size = data[nsi++].size;
                        }
                        else {
                            shifted = d.size >> 3;
                            decremented = (d.size - shifted < d.size) ? (d.size - shifted) : d.size - 1;
                            d.size = decremented;
                        }
                        if (~~d.size > 0) {
                            cloudSprite(contextAndRatio, d, data, i);
                        }
                    }
                }
            }
            if (i >= n || (start - stepStart) >= stepLimit) {
                cloud.stop();
                event.call('end', cloud, tags, bounds);
            }
        }
    };
    cloud.stop = function () {
        if (timer) {
            clearInterval(timer);
            timer = null;
        }
        return cloud;
    };
    function getContext(canvas) {
        canvas.width = canvas.height = 1;
        var ratio = Math.sqrt(canvas.getContext("2d").getImageData(0, 0, 1, 1).data.length >> 2);
        canvas.width = (cw << 5) / ratio;
        canvas.height = ch / ratio;
        var context = canvas.getContext("2d");
        context.fillStyle = context.strokeStyle = "red";
        context.textAlign = "center";
        return { context: context, ratio: ratio };
    }
    function place(board, tag, bounds) {
        var perimeter = [{ x: 0, y: 0 }, { x: size[0], y: size[1] }], startX = tag.x, startY = tag.y, maxDelta = Math.sqrt(size[0] * size[0] + size[1] * size[1]), s = spiral(size), dt = 1, t = -dt, dxdy, dx, dy;
        while (dxdy = s(t += dt)) {
            dx = ~~dxdy[0];
            dy = ~~dxdy[1];
            if (Math.min(Math.abs(dx), Math.abs(dy)) >= maxDelta)
                break;
            tag.x = startX + dx;
            tag.y = startY + dy;
            if (tag.x + tag.x0 < 0 || tag.y + tag.y0 < 0 ||
                tag.x + tag.x1 > size[0] || tag.y + tag.y1 > size[1])
                continue;
            // TODO only check for collisions within current bounds.
            if (!bounds || !cloudCollide(tag, board, size[0])) {
                if (!bounds || collideRects(tag, bounds)) {
                    var sprite = tag.sprite, w = tag.width >> 5, sw = size[0] >> 5, lx = tag.x - (w << 4), sx = lx & 0x7f, msx = 32 - sx, h = tag.y1 - tag.y0, x = (tag.y + tag.y0) * sw + (lx >> 5), last;
                    for (var j = 0; j < h; j++) {
                        last = 0;
                        for (var i = 0; i <= w; i++) {
                            if (sprite) {
                                board[x + i] |= (last << msx) | (i < w ? (last = sprite[j * w + i]) >>> sx : 0);
                            }
                        }
                        x += sw;
                    }
                    delete tag.sprite;
                    return true;
                }
            }
        }
        return false;
    }
    cloud.timeInterval = function (_) {
        return arguments.length ? (timeInterval = _ == null ? Infinity : _, cloud) : timeInterval;
    };
    cloud.stepLimit = function (_) {
        return arguments.length ? (stepLimit = _ == null ? Infinity : _, cloud) : stepLimit;
    };
    cloud.words = function (_) {
        return arguments.length ? (words = _, cloud) : words;
    };
    cloud.size = function (_) {
        return arguments.length ? (size = [+_[0], +_[1]], cloud) : size;
    };
    cloud.font = function (_) {
        return arguments.length ? (font = functor(_), cloud) : font;
    };
    cloud.fontStyle = function (_) {
        return arguments.length ? (fontStyle = functor(_), cloud) : fontStyle;
    };
    cloud.fontWeight = function (_) {
        return arguments.length ? (fontWeight = functor(_), cloud) : fontWeight;
    };
    cloud.rotate = function (_) {
        return arguments.length ? (rotate = functor(_), cloud) : rotate;
    };
    cloud.text = function (_) {
        return arguments.length ? (text = functor(_), cloud) : text;
    };
    cloud.spiral = function (_) {
        return arguments.length ? (spiral = spirals[_] || _, cloud) : spiral;
    };
    cloud.fontSize = function (_) {
        return arguments.length ? (fontSize = functor(_), cloud) : fontSize;
    };
    cloud.padding = function (_) {
        return arguments.length ? (padding = functor(_), cloud) : padding;
    };
    cloud.random = function (_) {
        return arguments.length ? (random = _, cloud) : random;
    };
    cloud.on = function () {
        var value = event.on.apply(event, arguments);
        return value === event ? cloud : value;
    };
    return cloud;
}
function cloudText(d) {
    return d.text;
}
function cloudFont() {
    return "serif";
}
function cloudFontNormal() {
    return "normal";
}
function cloudFontSize(d) {
    return Math.sqrt(d.value);
}
function cloudRotate() {
    return (~~(Math.random() * 6) - 3) * 30;
}
function cloudPadding() {
    return 1;
}
// Fetches a monochrome sprite bitmap for the specified text.
// Load in batches for speed.
function cloudSprite(contextAndRatio, d, data, di) {
    if (d.sprite)
        return;
    var c = contextAndRatio.context, ratio = contextAndRatio.ratio;
    c.clearRect(0, 0, (cw << 5) / ratio, ch / ratio);
    var x = 0, y = 0, maxh = 0, n = data.length;
    --di;
    while (++di < n) {
        d = data[di];
        c.save();
        c.font = d.style + " " + d.weight + " " + ~~((d.size + 1) / ratio) + "px " + d.font;
        var w = c.measureText(d.text + "m").width * ratio, h = d.size << 1;
        if (d.rotate) {
            var sr = Math.sin(d.rotate * cloudRadians), cr = Math.cos(d.rotate * cloudRadians), wcr = w * cr, wsr = w * sr, hcr = h * cr, hsr = h * sr;
            w = (Math.max(Math.abs(wcr + hsr), Math.abs(wcr - hsr)) + 0x1f) >> 5 << 5;
            h = ~~Math.max(Math.abs(wsr + hcr), Math.abs(wsr - hcr));
        }
        else {
            w = (w + 0x1f) >> 5 << 5;
        }
        if (h > maxh)
            maxh = h;
        if (x + w >= (cw << 5)) {
            x = 0;
            y += maxh;
            maxh = 0;
        }
        if (y + h >= ch)
            break;
        c.translate((x + (w >> 1)) / ratio, (y + (h >> 1)) / ratio);
        if (d.rotate)
            c.rotate(d.rotate * cloudRadians);
        c.fillText(d.text, 0, 0);
        if (d.padding)
            c.lineWidth = 2 * d.padding, c.strokeText(d.text, 0, 0);
        c.restore();
        d.width = w;
        d.height = h;
        d.xoff = x;
        d.yoff = y;
        d.x1 = w >> 1;
        d.y1 = h >> 1;
        d.x0 = -d.x1;
        d.y0 = -d.y1;
        d.hasText = true;
        x += w;
    }
    var pixels = c.getImageData(0, 0, (cw << 5) / ratio, ch / ratio).data, sprite = [];
    while (--di >= 0) {
        d = data[di];
        if (!d.hasText)
            continue;
        var w = d.width, w32 = w >> 5, h = d.y1 - d.y0;
        // Zero the buffer
        for (var i = 0; i < h * w32; i++)
            sprite[i] = 0;
        x = d.xoff;
        if (x == null)
            return;
        y = d.yoff;
        var seen = 0, seenRow = -1;
        for (var j = 0; j < h; j++) {
            for (var i = 0; i < w; i++) {
                var k = w32 * j + (i >> 5), m = pixels[((y + j) * (cw << 5) + (x + i)) << 2] ? 1 << (31 - (i % 32)) : 0;
                sprite[k] |= m;
                seen |= m;
            }
            if (seen)
                seenRow = j;
            else {
                d.y0++;
                h--;
                j--;
                y++;
            }
        }
        d.y1 = d.y0 + seenRow;
        d.sprite = sprite.slice(0, (d.y1 - d.y0) * w32);
    }
}
// Use mask-based collision detection.
function cloudCollide(tag, board, sw) {
    sw >>= 5;
    var sprite = tag.sprite, w = tag.width >> 5, lx = tag.x - (w << 4), sx = lx & 0x7f, msx = 32 - sx, h = tag.y1 - tag.y0, x = (tag.y + tag.y0) * sw + (lx >> 5), last;
    for (var j = 0; j < h; j++) {
        last = 0;
        for (var i = 0; i <= w; i++) {
            if (sprite && ((last << msx) | (i < w ? (last = sprite[j * w + i]) >>> sx : 0))
                & board[x + i])
                return true;
        }
        x += sw;
    }
    return false;
}
function cloudBounds(bounds, d) {
    var b0 = bounds[0], b1 = bounds[1];
    if (d.x + d.x0 < b0.x)
        b0.x = d.x + d.x0;
    if (d.y + d.y0 < b0.y)
        b0.y = d.y + d.y0;
    if (d.x + d.x1 > b1.x)
        b1.x = d.x + d.x1;
    if (d.y + d.y1 > b1.y)
        b1.y = d.y + d.y1;
}
function collideRects(a, b) {
    return a.x + a.x1 > b[0].x && a.x + a.x0 < b[1].x && a.y + a.y1 > b[0].y && a.y + a.y0 < b[1].y;
}
function archimedeanSpiral(size) {
    var e = size[0] / size[1];
    return function (t) {
        return [e * (t *= .1) * Math.cos(t), t * Math.sin(t)];
    };
}
function rectangularSpiral(size) {
    var dy = 4, dx = dy * size[0] / size[1], x = 0, y = 0;
    return function (t) {
        var sign = t < 0 ? -1 : 1;
        // See triangular numbers: T_n = n * (n + 1) / 2.
        switch ((Math.sqrt(1 + 4 * sign * t) - sign) & 3) {
            case 0:
                x += dx;
                break;
            case 1:
                y += dy;
                break;
            case 2:
                x -= dx;
                break;
            default:
                y -= dy;
                break;
        }
        return [x, y];
    };
}
// TODO reuse arrays?
function zeroArray(n) {
    var a = [], i = -1;
    while (++i < n)
        a[i] = 0;
    return a;
}
function cloudCanvas() {
    return document.createElement("canvas");
}
function functor(d) {
    return typeof d === "function" ? d : function () { return d; };
}
var spirals = {
    archimedean: archimedeanSpiral,
    rectangular: rectangularSpiral
};

function useResponsiveSVGSelection(minSize, initialSize) {
    var elementRef = useRef();
    var _a = useState(initialSize), size = _a[0], setSize = _a[1];
    var _b = useState(null), selection = _b[0], setSelection = _b[1];
    useEffect(function () {
        var element = elementRef.current;
        // set svg selection
        var svg = select(element)
            .append('svg')
            .style('display', 'block'); // native inline svg leaves undesired white space
        var selection = svg.append('g');
        setSelection(selection);
        function updateSize(width, height) {
            svg.attr('height', height).attr('width', width);
            selection.attr('transform', "translate(" + width / 2 + ", " + height / 2 + ")");
            setSize([width, height]);
        }
        var width = 0;
        var height = 0;
        if (initialSize !== undefined) {
            // Use initialSize if it is provided
            width = initialSize[0], height = initialSize[1];
        }
        else {
            // Use parentNode size if resized has not updated
            width = element.parentElement.offsetWidth;
            height = element.parentElement.offsetHeight;
        }
        width = Math.max(width, minSize[0]);
        height = Math.max(height, minSize[1]);
        updateSize(width, height);
        // update resize using a resize observer
        var resizeObserver = new ResizeObserver(function (entries) {
            if (!entries || !entries.length) {
                return;
            }
            if (initialSize === undefined) {
                var _a = entries[0].contentRect, width_1 = _a.width, height_1 = _a.height;
                updateSize(width_1, height_1);
            }
        });
        resizeObserver.observe(element);
        // cleanup
        return function () {
            resizeObserver.unobserve(element);
            select(element)
                .selectAll('*')
                .remove();
        };
    }, [initialSize[0], initialSize[1], minSize[0], minSize[1]]);
    return [elementRef, selection, size];
}

var Scale;
(function (Scale) {
    Scale["Linear"] = "linear";
    Scale["Log"] = "log";
    Scale["Sqrt"] = "sqrt";
})(Scale || (Scale = {}));
var Spiral;
(function (Spiral) {
    Spiral["Archimedean"] = "archimedean";
    Spiral["Rectangular"] = "rectangular";
})(Spiral || (Spiral = {}));

function choose(array, random) {
    return array[Math.floor(random() * array.length)];
}
function getDefaultColors() {
    return range(20)
        .map(function (number) { return number.toString(); })
        .map(scaleOrdinal(schemeCategory10));
}
function getText(word) {
    return word.text;
}
function getFontSize(word) {
    return word.size + "px";
}
function getTransform(word) {
    var translate = "translate(" + word.x + ", " + word.y + ")";
    var rotate = typeof word.rotate === 'number' ? "rotate(" + word.rotate + ")" : '';
    return translate + rotate;
}
function rotate(rotations, rotationAngles, random) {
    if (rotations < 1) {
        return 0;
    }
    var angles = [];
    if (rotations === 1) {
        angles = [rotationAngles[0]];
    }
    else {
        angles = __spreadArrays(rotationAngles);
        var increment = (rotationAngles[1] - rotationAngles[0]) / (rotations - 1);
        var angle = rotationAngles[0] + increment;
        while (angle < rotationAngles[1]) {
            angles.push(angle);
            angle += increment;
        }
    }
    return choose(angles, random);
}

var tooltipInstance;
function render(selection, words, options, callbacks, random) {
    var getWordColor = callbacks.getWordColor, getWordTooltip = callbacks.getWordTooltip, onWordClick = callbacks.onWordClick, onWordMouseOver = callbacks.onWordMouseOver, onWordMouseOut = callbacks.onWordMouseOut, onRenderComplete = callbacks.onRenderComplete;
    var colors = options.colors, enableTooltip = options.enableTooltip, fontStyle = options.fontStyle, fontWeight = options.fontWeight;
    var fontFamily = options.fontFamily, transitionDuration = options.transitionDuration;
    function getFill(word, index) {
        return getWordColor ? getWordColor(word, index) : choose(colors, random);
    }
    // load words
    var vizWords = selection.selectAll('text').data(words);
    vizWords.join(function (enter) {
        return enter
            .append('text')
            .on('click', function (word) {
            onWordClick && onWordClick(word, event);
        })
            .on('mouseover', function (word) {
            if (enableTooltip) {
                tooltipInstance = tippy(event.target, {
                    animation: 'scale',
                    arrow: true,
                    content: function () {
                        return getWordTooltip(word);
                    },
                });
            }
            onWordMouseOver && onWordMouseOver(word, event);
        })
            .on('mouseout', function (word) {
            if (tooltipInstance) {
                tooltipInstance.destroy();
            }
            onWordMouseOut && onWordMouseOut(word, event);
        })
            .attr('cursor', onWordClick ? 'pointer' : 'default')
            .attr('fill', getFill)
            .attr('font-family', fontFamily)
            .attr('font-style', fontStyle)
            .attr('font-weight', fontWeight)
            .attr('text-anchor', 'middle')
            .attr('transform', 'translate(0, 0) rotate(0)')
            .call(function (enter) {
            return enter
                .transition()
                .duration(transitionDuration)
                .attr('font-size', getFontSize)
                .attr('transform', getTransform)
                .text(getText);
        });
    }, 
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    function (update) {
        return update
            .transition()
            .duration(transitionDuration)
            .attr('fill', getFill)
            .attr('font-family', fontFamily)
            .attr('font-size', getFontSize)
            .attr('transform', getTransform)
            .text(getText);
    }, function (exit) {
        exit
            .remove();
    });
    if (onRenderComplete) {
        onRenderComplete();
    }
}

var defaultCallbacks = {
    getWordTooltip: function (_a) {
        var text = _a.text, value = _a.value;
        return text + " (" + value + ")";
    },
};
var defaultOptions = {
    colors: getDefaultColors(),
    deterministic: false,
    enableTooltip: true,
    fontFamily: 'times new roman',
    fontSizes: [4, 32],
    fontStyle: 'normal',
    fontWeight: 'normal',
    padding: 1,
    rotationAngles: [-90, 90],
    scale: Scale.Sqrt,
    spiral: Spiral.Rectangular,
    transitionDuration: 600,
};
function Wordcloud(_a) {
    var callbacks = _a.callbacks, _b = _a.maxWords, maxWords = _b === void 0 ? 100 : _b, minSize = _a.minSize, options = _a.options, initialSize = _a.size, words = _a.words;
    // without useMemo the component re-renders on every upstream change
    // because of the creation of new mergedCallbacks and mergedOptions variables
    var mergedCallbacks = useMemo(function () { return __assign(__assign({}, defaultCallbacks), callbacks); }, []);
    var mergedOptions = useMemo(function () { return __assign(__assign({}, defaultOptions), options); }, []);
    var _c = useResponsiveSVGSelection(minSize, initialSize), ref = _c[0], selection = _c[1], size = _c[2];
    useEffect(function () {
        if (selection) {
            var deterministic = mergedOptions.deterministic, fontFamily = mergedOptions.fontFamily, fontStyle = mergedOptions.fontStyle, fontSizes = mergedOptions.fontSizes, fontWeight = mergedOptions.fontWeight, padding = mergedOptions.padding, rotations_1 = mergedOptions.rotations, rotationAngles_1 = mergedOptions.rotationAngles, spiral = mergedOptions.spiral, stepLimit = mergedOptions.stepLimit, timeInterval = mergedOptions.timeInterval, scale = mergedOptions.scale;
            var sortedWords = words
                .concat()
                .sort(function (x, y) { return descending(x.value, y.value); })
                .slice(0, maxWords);
            var random_1 = deterministic ? seedrandom('deterministic') : seedrandom();
            var layout_1 = cloud()
                .size(size)
                .padding(padding)
                .rotate(function () {
                if (rotations_1 === undefined) {
                    // default rotation algorithm
                    return (~~(random_1() * 6) - 3) * 30;
                }
                else {
                    return rotate(rotations_1, rotationAngles_1, random_1);
                }
            })
                .spiral(spiral)
                .stepLimit(stepLimit)
                .timeInterval(timeInterval)
                .random(random_1)
                .text(getText)
                .font(fontFamily)
                .fontStyle(fontStyle)
                .fontWeight(fontWeight)
                .stepLimit(stepLimit)
                .timeInterval(timeInterval);
            var draw = function (fontSizes, words, attempts) {
                layout_1
                    .words(words)
                    .fontSize(function (word) {
                    // just returning the value because we apply our own font scaling ahead of time
                    return word.value;
                })
                    .on('end', function (computedWords) {
                    render(selection, computedWords, mergedOptions, mergedCallbacks, random_1);
                })
                    .start();
            };
            draw(fontSizes, sortedWords);
        }
    }, [maxWords, mergedCallbacks, mergedOptions, selection, size, words]);
    return React.createElement("div", { ref: ref });
}
Wordcloud.defaultProps = {
    callbacks: defaultCallbacks,
    maxWords: 100,
    minSize: [300, 300],
    options: defaultOptions,
};

export default Wordcloud;
export { Scale, Spiral, defaultCallbacks, defaultOptions };
